# Copyright 2011 Citrix Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from debtcollector import removals

from neutron.pecan_wsgi import app as pecan_app


class Versions(object):

    @removals.remove(version="Queens", removal_version="Rocky",
                     message="Use neutron.pecan_wsgi.app:versions_factory")
    @classmethod
    def factory(cls, global_config, **local_config):
        return pecan_app.versions_factory(global_config, **local_config)
